#-*- coding: utf-8 -*-
from distutils.core import setup
setup(
  name='pydepr',
  packages=['pydepr'],
  version='0.14',
  description='PyDePr is a set of tools for processing machine learning models.',
  author='Eric Strong',
  author_email='dr.eric.strong@gmail.com',
  url='https://github.com/drericstrong/pydepr',
  download_url='https://github.com/drericstrong/pydepr/tarball/0.13',
  keywords=['regression', 'data'],
  classifiers=[],
)
